#! /usr/bin/python3

import aquests
import os, sys

def request_finished_and_req (r):
	global TOTAL, DONE
	DONE += 1
	if (DONE + len (aquests._currents)) < TOTAL:
		aquests.get (r.url)
	args = (r.status_code, r.reason, len (r.content))
	print ("%s %s %d bytes received" % args, r.version)

def request_finished (r):
	args = (r.status_code, r.reason, len (r.content))
	print ("%s %s %d bytes received" % args, r.version)


def make_example_testset ():
	jpg = open (os.path.join (r"D:\apps\skitai\tests\examples", "statics", "reindeer.jpg"), "rb")	
	aquests.upload ("http://127.0.0.1:5000/upload", {"username": "pytest", "file1": jpg})		
	aquests.get ("http://127.0.0.1:5000/")	
	aquests.get ("http://127.0.0.1:5000/members/", auth = ('admin', '1111'))
	aquests.get ("http://127.0.0.1:5000/redirect1")	
	aquests.rpc ("http://127.0.0.1:5000/rpc2/").add_number (1, 2)	
	aquests.ws ("ws://127.0.0.1:5000/websocket/echo-single", "Hello Skitai")
	
def test_load (h1, client, streams, urls):
	if urls == 1:
		aquests.configure (client, callback = request_finished, force_http1 = h1, http2_constreams = streams)
		for i in range (TOTAL):
			make_example_testset ()				
	else:
		aquests.configure (client, callback = request_finished_and_req, force_http1 = h1, http2_constreams = streams)			
		for i in range (client * streams):
			for url in urls:
				aquests.get (url)

	aquests.fetchall ()

def usage ():
	print ("""
python -m aquests.load [options] urls
ex) python -m aquests.load -n 100 -c2 -m2 http://yourserver.com

options:
  
  -n: number of requests
  -c: number of clients
  -m: numner of HTTP2 concurent streams
  --h1: use force HTTP/1.1 request, -m option will be ignored  
	""")
	sys.exit ()

if __name__ == "__main__":
	import getopt, time
	argopt = getopt.getopt(sys.argv[1:], "en:c:m:", ["h1", "help"])
	
	h1 = 0
	num = 1
	client = 1
	streams = 1
	test_example = 0
	TOTAL = 0
	DONE = 0
	for k, v in argopt [0]:
		if k == "--h1":
			h1 = 1
		elif k == "--help":
			usage ()
		elif k == "-n":
			TOTAL = int (v)
		elif k == "-c":
			client = int (v)
		elif k == "-m":
			streams = int (v)
		elif k == "-e":	
			test_example = 1
			
	test_load (h1, min (client, TOTAL), streams, test_example or argopt [1])
	


from . import request_handler as http_request_handler
from . import response

DEFAULT_USER_AGENT = "Mozilla/5.0 (compatible; Skitaibot/0.1a)"	

class ProxyTunnelHandler (http_request_handler.RequestHandler):
	def __init__ (self, asyncon, request, callback, *args, **karg):
		http_request_handler.RequestHandler.__init__ (self, asyncon, request, callback, connection = "keep-alive")
	
	def get_handshaking_buffer (self):
		req = ("CONNECT %s:%d HTTP/1.1\r\nUser-Agent: %s\r\n\r\n" % (
					self.request.address [0], 
					self.request.address [1],					
					DEFAULT_USER_AGENT
		)).encode ("utf8")
		return req
	
	def start_handshake (self):
		self.asyncon.set_terminator ("\r\n\r\n")
		self.asyncon.push (self.get_handshaking_buffer ())
		self.asyncon.begin_tran (self)
	
	def convert_to_ssl (self):
		pass
		
	def finish_handshake (self):
		if self.response.code == 200:
			self.asyncon.established = True
			self.response = None
			self.convert_to_ssl ()
			for buf in self.get_request_buffer ():
				self.asyncon.push (buf)
		else:
			self.response = response.FailedResponse (self.response.code, self.response.msg)
			self.asyncon.handle_close (720, "Returned %d %s" % (self.response.code, self.response.msg))
							
	def handle_request (self):	
		if not self.asyncon.established:
			self.start_handshake ()
		else:
			http_request_handler.RequestHandler.handle_request (self)
		
	def found_end_of_body (self):			
		if not self.asyncon.established:
			self.finish_handshake ()											
		else:
			http_request_handler.RequestHandler.found_end_of_body (self)
		
	def connection_closed (self, why, msg):
		if self.response is None:
			self.response = response.FailedResponse (why, msg)
		self.close_case ()


class SSLProxyTunnelHandler (ProxyTunnelHandler):
	def convert_to_ssl (self):
		self.asyncon.established = True
		self.asyncon.connected = False
		self.asyncon.connecting = True




from ..protocols.dns import asyndns
import time

class DNSCache:
	def __init__ (self, logger = None, dns_servers = []):		
		self.dns = asyndns.Request (logger)
		self.logger = logger
		self.dns_servers = dns_servers
		self.cache = {}
		self.hits = 0

	def set (self, answers):
		if not answers:
			return

		for answer in answers:
			name = answer ["name"].lower ()
			if name not in self.cache:
				self.cache [name] = {}
			
			if not answer ["data"]:
				addrs = self.get (name, answer ["typename"], False)
				if addrs and addrs [0]["data"]:
					return
						
			if "ttl" in answer:
				answer ["valid"]	= time.time () + answer ["ttl"]
			self.cache [name][answer ["typename"]] = [answer]
		
	def expire (self, host):
		try: del self.cache [host]
		except KeyError: pass		

	def get (self, host, qtype = "A", check_ttl = True):	
		host = host.lower ()
		try: answers = self.cache [host][qtype]
		except KeyError: return []
		
		answer = answers [0]
		if "valid" not in answer:
			return [answer]
			
		else:
			now = time.time ()
			if check_ttl and answer ["valid"] < now:				
				# use max 5 minutes seconds for other querees
				answer ['valid'] = now + (answer ["data"] and 300 or 1)
				# nut new query will be started				
				return []
				
			else:
				return [answer]
	
	def is_ip (self, name):
		arr = name.split (".")
		if len (arr) != 4: return False
		try: arr = [x for x in map (int, arr) if x & 255 == x]
		except ValueError: 
			return False
		if len (arr) != 4: return False
		return True
		
	def __call__ (self, host, qtype, callback):
		self.hits += 1
		hit = self.get (host, qtype, True)		
		if hit: 
			return callback (hit)
	
		if self.is_ip (host):
			self.set ([{"name": host, "data": host, "typename": qtype}])
			return callback ([{"name": host, "data": host, "typename": qtype}])
		
		try:
			self.dns.req (host, server = self.dns_servers, qtype = qtype, protocol = "tcp", callback = [self.set, callback])
		except:
			self.logger.trace (host)
			hit = [{"name": host, "data": None, "typename": qtype, 'ttl': 60}]
			self.set (hit)
			callback (hit)


PUBLIC_DNS_SERVERS = ['8.8.8.8', '8.8.4.4']
query = None

def init (logger, dns_servers = []):
	if not dns_servers:
		dns_servers = PUBLIC_DNS_SERVERS
	global query
	if query is None:
		query = DNSCache (logger, dns_servers)	

def get (name, qtype = "A"):	
	global query
	return query.get (name, qtype, False)
